package os.file;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import os.io.FileInputService;
import os.io.FileOutputService;

/**
 * Manager for working with file.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 * 
 */
public class FileManager {

    private static Logger logger = Logger.getLogger(FileManager.class.getName());
    private static FileManager manager;

    private String rootDir;
    private String helpDir;
    private String historyDir;

    public FileManager() {

        // get actual directory where app run
        File directory = new File(".");
        String absolutePath = directory.getAbsolutePath();
        rootDir = getPathBasedOnOS(absolutePath.subSequence(0, absolutePath.length() - 2).toString());

        // create help dir if doesnt exist
        helpDir = getDefaultDir() + File.separator + "help";

        File helpDirectory = new File(helpDir);
        if (!helpDirectory.exists())
            helpDirectory.mkdir();

        // create history dir if doesnt exist
        historyDir = getDefaultDir() + File.separator + "history";
        File historyDirectory = new File(historyDir);
        if (!historyDirectory.exists())
            historyDirectory.mkdir();

    }

    public static FileManager get() {

        if (manager == null)
            init();

        return manager;
    }

    private static void init() {

        logger.log(Level.INFO, "initializing File manager \n");
        manager = new FileManager();
        logger.log(Level.INFO, "initialized File manager \n");
    }
    
    /**
     * Shutdown file manager.
     */
    public void shutdown() {
        logger.log(Level.INFO, "shutdown File manager \n");
    }
    
    /**
     * Get file or null from path.
     * 
     */
    public File getFile(String absolutePath) {

        File file = new File(absolutePath);

        if (file.exists() && !file.isDirectory()) {
            return file;
        }

        return null;
    }
    
    /**
     * Get file or dir or null from path.
     * 
     */
    public File getFileOrDir(String absolutePath) {

        File file = new File(absolutePath);

        if (file.exists()) {
            return file;
        }

        return null;
    }
    
    /**
     * Get file or dir or null from path and fileName.
     * 
     */
    public File getFileOrDir(String path, String fileName) {

        File file = new File(path, fileName);

        if (file.exists()) {
            return file;
        }

        return null;
    }
    
    /**
     * Get file or dir or null from path and fileName.
     * 
     */
    public File getFileOrDir(File path, String fileName) {

        File file = new File(path, fileName);

        if (file.exists()) {
            return file;
        }

        return null;
    }

    /**
     * Get file or dir or null from path.
     * 
     */
    public File getOrCreateFile(String absolutePath) {

        File file = new File(absolutePath);
        if (file.exists()) {
            return file;
        } else {
            try {

                file.createNewFile();
                return file;

            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                return null;
            }
        }
    }
    
    /**
     * Get or create file from path and fileName.
     * 
     */
    public File getOrCreateFile(File path, String fileName) {

        File file = new File(path, fileName);
        if (file.exists()) {
            return file;
        } else {
            try {

                file.createNewFile();
                return file;

            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                return null;
            }
        }
    }
    
    /**
     * Get or create file from path and fileName.
     * 
     */
    public File getOrCreateFile(String path, String fileName) {

        File file = new File(path, fileName);
        if (file.exists()) {
            return file;
        } else {
            try {

                file.createNewFile();
                return file;

            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                return null;
            }
        }
    }

    public String getRootDir() {
        return rootDir;
    }

    public String getDefaultDir() {
        return getRootDir();
    }
    
    
    /**
     * Get list of all files/folders from directory pathname.
     * 
     */
    public File[] getListOfFiles(String pathname) {

        File dir = new File(pathname);

        if (dir.isDirectory()) {
            return dir.listFiles();
        }

        return null;
    }

    public boolean isDir(String pathname) {

        File dir = new File(pathname);

        if (dir.exists()) {
            return dir.isDirectory();
        }

        return false;
    }

    public boolean isAbsoluteDir(String pathname) {

        File dir = new File(pathname);

        if (dir.exists()) {
            return (dir.isDirectory() && dir.isAbsolute());
        }

        return false;
    }

    public boolean isPath(String pathname) {

        File path = new File(pathname);
        return path.exists();
    }

    public String getHelpDir() {
        return helpDir;
    }

    public String getHistoryDir() {
        return historyDir;
    }
    
    /**
     * Get or create dir from path and fileName.
     * 
     */
    public File getOrCreateDir(File path, String fileName) {

        File file = new File(path, fileName);
        if (file.exists()) {
            return file;
        } else {
            file.mkdir();
            return file;
        }
    }
    
    /**
     * Get or create dir from path.
     * 
     */
    public File getOrCreateDir(String path) {

        File file = new File(path);
        if (file.exists()) {
            return file;
        } else {
            file.mkdir();
            return file;
        }
    }
    
    /**
     * Get system based path from path.
     * 
     */
    public String getPathBasedOnOS(String path) {

        try {

            String newPath = path.replace('\\', '/');

            if (newPath.contains("//")) {
                return null;
            }

            newPath.replace('/', File.separatorChar);

            File file = new File(newPath);

            return file.getCanonicalPath();

        } catch (IOException e) {
            return null;
        }
    }
    
    /**
     * Open file on path and fileName and return this file like input service.
     */
    public FileInputService openFileAsInput(String path, String fileName) throws FileNotFoundException {

        String filePath = null;

        if (getFile(getPathBasedOnOS(fileName)) == null) {

            if (!path.endsWith(File.separator)) {
                filePath = getPathBasedOnOS(path + File.separator + fileName);
            } else {
                filePath = getPathBasedOnOS(path + fileName);
            }

        } else {
            filePath = getPathBasedOnOS(fileName);
        }

        if (filePath != null) {

            File inputFile = getFile(filePath);
            if (inputFile != null) {
                return new FileInputService(inputFile);
            } else {
                return null;
            }

        } else {
            return null;
        }
    }
    
    /**
     * Open file on path and fileName and return this file like output service.
     */
    public FileOutputService openFileAsOutput(String path, String fileName, boolean errorOutput, boolean append) throws IOException {

        String filePath = null;

        if (!path.endsWith(File.separator)) {
            filePath = getPathBasedOnOS(path + File.separator + fileName);
        } else {
            filePath = getPathBasedOnOS(path + fileName);
        }

        if (filePath != null) {

            File outputFile = getOrCreateFile(filePath);

            if (outputFile != null) {
                return new FileOutputService(outputFile, errorOutput, append);
            } else {
                return null;
            }

        } else {
            return null;
        }
    }
}
