package os.history;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;
import java.util.logging.Logger;

import os.file.FileManager;
import os.session.Session;

/**
 * Manager for management of history.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 * 
 */
public class HistoryManager {

    private static Logger logger = Logger.getLogger(FileManager.class.getName());

    private static HistoryManager manager;
    private static final int HISTORY_MAX_COMMANDS = 100;
    private static final String HISTORY_FILE_PREFIX = "history_";
    private static final String HISTORY_FILE_SUFIX = ".hist";
    private static final String HISTORY_ITEM_SEPARATOR = ";";
    private static final String ENDLINE = "\n";

    private Map<String, History> historyMap = new HashMap<String, History>();
    private ReentrantLock lock = new ReentrantLock();

    private static void init() {

        logger.log(Level.INFO, "initializing History manager \n");
        manager = new HistoryManager();
        logger.log(Level.INFO, "initialized History manager \n");
    }

    public static HistoryManager get() {

        if (manager == null)
            init();

        return manager;
    }

    public HistoryManager() {

    }

    /**
     * Shutdown history manager.
     */
    public void shutdown() {
        logger.log(Level.INFO, "shutdown History manager \n");
    }

    /**
     * Try to find history for session.
     */
    public History getHistoryForSession(Session session) {

        logger.log(Level.INFO, "loading history for session " + session.getLogin());
        List<HistoryItem> list = tryLoadHistoryFile(getHistoryFileName(session.getLogin()));

        // history not exist
        if (list == null) {
            lock.lock();
            historyMap.put(session.getLogin(), new History(session.getLogin()));
            lock.unlock();
        } else {
            // history exist
            History history = new History(session.getLogin());
            history.addFullHistory(list);

            lock.lock();
            historyMap.put(session.getLogin(), history);
            lock.unlock();
        }

        lock.lock();
        History history = historyMap.get(session.getLogin());
        lock.unlock();

        return history;
    }

    /**
     * Load history from file if exist.
     */
    private List<HistoryItem> tryLoadHistoryFile(String historyFileName) {

        File historyFile = FileManager.get().getFileOrDir(FileManager.get().getHistoryDir(), historyFileName);

        if (historyFile == null) {
            return null;
        } else {

            try {

                InputStreamReader input;
                input = new InputStreamReader(new FileInputStream(historyFile), "UTF8");
                BufferedReader reader = new BufferedReader(input);
                List<HistoryItem> history = new ArrayList<HistoryItem>();

                String line;

                while ((line = reader.readLine()) != null && !line.isEmpty()) {

                    String[] values = line.split(HISTORY_ITEM_SEPARATOR);
                    if (values.length == 2) {
                        history.add(new HistoryItem(values[0], values[1]));
                    }
                }

                reader.close();
                input.close();

                return history;

            } catch (UnsupportedEncodingException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            } catch (FileNotFoundException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }

            return null;
        }
    }

    /**
     * Save history in file from session.
     */
    public void saveHistory(Session session) {

        logger.log(Level.INFO, "saving history for session " + session.getLogin());
        File historyFile = FileManager.get().getOrCreateFile(FileManager.get().getHistoryDir(), getHistoryFileName(session.getLogin()));

        try {

            OutputStreamWriter output = new OutputStreamWriter(new FileOutputStream(historyFile, false), "UTF8");
            BufferedWriter writer = new BufferedWriter(output);

            lock.lock();
            History history = historyMap.get(session.getLogin());

            if (history == null)
                history = session.getHistory();

            historyMap.remove(session.getLogin());
            lock.unlock();

            List<HistoryItem> historyList = history.getHistory();
            if (historyList.size() > HISTORY_MAX_COMMANDS) {
                historyList = historyList.subList(historyList.size() - 1 - HISTORY_MAX_COMMANDS, historyList.size() - 1);
            }

            for (HistoryItem item : historyList) {

                writer.write(item.getDir() + HISTORY_ITEM_SEPARATOR + item.getCommand() + ENDLINE);
            }

            writer.close();
            output.close();

        } catch (UnsupportedEncodingException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        } catch (FileNotFoundException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
    }

    private String getHistoryFileName(String login) {
        return HISTORY_FILE_PREFIX + login + HISTORY_FILE_SUFIX;
    }
}
