package os.io;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.logging.Level;
import java.util.logging.Logger;

import os.session.Session;

/**
 * Input/Output service for piped processes.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 * 
 */
public class PipeInputOutputService implements InputService, OutputService {

    private static Logger logger = Logger.getLogger(PipeInputOutputService.class.getName());

    private BufferedWriter output;
    private BufferedReader input;
    private PipedOutputStream outputPipe;
    private PipedInputStream inputPipe;

    public PipeInputOutputService() {

        try {

            // create input and output pipe, input pipe has size 16384 B
            outputPipe = new PipedOutputStream();
            inputPipe = new PipedInputStream(16384);

            // connect input pipe to output pipe
            inputPipe.connect(outputPipe);

            output = new BufferedWriter(new OutputStreamWriter(outputPipe, OutputService.CHARSET));
            input = new BufferedReader(new InputStreamReader(inputPipe, InputService.CHARSET));

        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
    }

    public void write(String text) {

        try {

            output.write(text);
            output.flush();

        } catch (IOException e) {
            // nothing log - IOException is throwed in kill process action and is not error.
            close();
        }
    }

    public void write(int c) {

        try {

            output.write(c);
            output.flush();

        } catch (IOException e) {
            // nothing log - IOException is throwed in kill process action and is not error.
            close();
        }
    }

    public void write(byte[] bytes) {

        try {

            String text = new String(bytes);
            output.write(text);
            output.flush();

        } catch (IOException e) {
            // nothing log - IOException is throwed in kill process action and is not error.
            close();
        }
    }

    public boolean isErrorService() {
        return false;
    }

    public String readLine() {

        try {

            return input.readLine();

        } catch (IOException e) {
            // nothing log - IOException is throwed in kill process action and is not error.
            close();
            return null;
        }
    }

    public int read() {

        try {

            return input.read();

        } catch (IOException e) {
            // nothing log - IOException is throwed in kill process action and is not error.
            close();
            return -1;
        }
    }

    public int read(byte[] bytes) {
        // pipe input side dont work via bytes
        throw new UnsupportedOperationException("read(byte[] bytes) cannot be used for read piped input service.");
    }

    public void close() {

        if (output != null) {
            try {

                output.flush();
                output.close();
            } catch (IOException e) {
                // output is closed
            }
        }
    }

    public int getType() {
        return InputService.PIPE;
    }

    public void setSession(Session session) {
        // nothing to do - pipe service doesn't need session.
    }

    public void shutdown() {
        // nothing to do.
    }
}
