package os.process;

import java.util.ArrayList;
import java.util.List;

import os.io.InputService;
import os.io.OutputService;
import os.process.shell.ShellProcess;
import os.process.shell.cat.CatProcess;
import os.process.shell.cd.CdProcess;
import os.process.shell.cp.CopyProcess;
import os.process.shell.date.DateProcess;
import os.process.shell.echo.EchoProcess;
import os.process.shell.exit.ExitProcess;
import os.process.shell.help.HelpProcess;
import os.process.shell.kill.KillProcess;
import os.process.shell.login.LoginProcess;
import os.process.shell.ls.LsProcess;
import os.process.shell.ps.PSProcess;
import os.process.shell.pwd.PwdProcess;
import os.process.shell.sort.SortProcess;
import os.session.Session;

/**
 * Abstract class for processes.
 * 
 * Base structure of process - input, output, error, parent process, childs processes.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 * 
 */
abstract public class Process implements IProcess, Comparable<Process> {

    /* Static list of all runnable programs for syntax analyzator */
    public static List<Process> runnable = new ArrayList<Process>() {

        private static final long serialVersionUID = 1L;

        {
            add(new CatProcess());
            add(new CdProcess());
            add(new CopyProcess());
            add(new DateProcess());
            add(new EchoProcess());
            add(new ExitProcess());
            add(new HelpProcess());
            add(new KillProcess());
            add(new LoginProcess());
            add(new LsProcess());
            add(new PSProcess());
            add(new PwdProcess());
            add(new ShellProcess());
            add(new ShutdownProcess());
            add(new SortProcess());
        }
    };

    private InputService input;
    private OutputService output;
    private OutputService error;

    private List<String> args = new ArrayList<String>();
    private List<String> params = new ArrayList<String>();

    private Long pid;
    private Long parent;
    private Long next;
    private List<Long> childs = new ArrayList<Long>();
    private List<Process> pipedProcesses = null;

    private boolean background;
    private boolean shutdown;
    private boolean logged;

    private Session session;

    public void setInput(InputService input) {
        this.input = input;
        getSesionToIO(input);

    }

    public void setOutput(OutputService output) {
        this.output = output;
        getSesionToIO(input);

    }

    public void setError(OutputService error) {
        this.error = error;

    }

    public Session getSession() {
        return this.session;
    }

    public void setSession(Session session) {
        this.session = session;
        getSesionToIO(input);

    }

    public void setArguments(List<String> args) {
        this.args = args;
    }

    public void setParameters(List<String> params) {
        this.params = params;
    }

    public void setParent(Long parent) {
        this.parent = parent;
    }

    public void addChild(Long child) {

        if (childs != null) {
            childs.add(child);
        }
    }

    public void setNext(Long next) {

        int index = childs.indexOf(next) + 1;
        if (index >= childs.size()) {
            this.next = null;
        } else {
            this.next = childs.get(index);
        }
    }

    public void setChilds(List<Long> childs) {
        this.childs = childs;
    }

    public void addArgument(String arg) {
        args.add(arg);
    }

    public void addParameter(String param) {
        params.add(param);
    }

    public List<String> getArgs() {
        return args;
    }

    public List<String> getParams() {
        return params;
    }

    public Long getParent() {
        return parent;
    }

    public Long getNext() {
        return next;
    }

    public List<Long> getChilds() {
        return childs;
    }

    public InputService getInput() {
        return input;
    }

    public OutputService getOutput() {
        return output;
    }

    public OutputService getError() {
        return error;
    }

    public Long getPid() {
        return pid;
    }

    public void setPid(Long pid) {
        this.pid = pid;
    }

    public void close() {

        if (input != null) {
            input.close();
        }

        if (output != null) {
            output.close();
        }

        if (error != null) {
            error.close();
        }

        if (background) {
            ProcessManager.get().endBackground(pid);
        }

    }

    /**
     * Go through all process and return yes/no if command is in list of runnable processes.
     * 
     */
    public boolean isProcess(String command) {

        for (Process runableProcess : runnable) {

            if (runableProcess.getCommand() != null && command.equalsIgnoreCase(runableProcess.getCommand())) {
                return true;
            }
        }

        return false;
    }

    public boolean isBackground() {
        return background;
    }

    public void setBackground(boolean background) {
        this.background = background;
    }

    public boolean isShutdown() {
        return shutdown;
    }

    public void setShutdown(boolean shutdown) {
        this.shutdown = shutdown;
    }

    public boolean isLogged() {
        return logged;
    }

    public void setLogged(boolean logged) {
        this.logged = logged;
    }

    public void shutdown() {
        setShutdown(true);
        getOutput().shutdown();
    }

    private void getSesionToIO(InputService input) {

        if (input != null && output != null) {
            input.setSession(session);
            output.setSession(session);
        }
    }

    public void setPipedProcesses(List<Process> pipedProcesses) {
        this.pipedProcesses = pipedProcesses;
    }

    public List<Process> getPipedProcesses() {
        return pipedProcesses;
    }

    public int compareTo(Process o) {
        // desc order
        return -1 * pid.compareTo(o.getPid());
    }
}
