package os.process;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;
import java.util.logging.Logger;

import os.io.EmptyOutputService;
import os.process.shell.ShellProcess;
import os.session.Session;

/**
 * Process manager for working and sheduling processes.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 * 
 */
public class ProcessManager {

    private static Logger logger = Logger.getLogger(ProcessManager.class.getName());
    private static ProcessManager manager;
    private static Long PIDGenerator = new Long(1);

    private ReentrantLock lock = new ReentrantLock();
    private Map<Long, Thread> threadMap = new HashMap<Long, Thread>();
    private Map<Long, IProcess> runnableMap = new HashMap<Long, IProcess>();
    private Map<Long, Long> sessionPidMap = new HashMap<Long, Long>();

    public ProcessManager() {

    }

    public static ProcessManager get() {

        if (manager == null)
            init();

        return manager;
    }

    private static void init() {
        logger.log(Level.INFO, "initializing Process manager \n");
        manager = new ProcessManager();
        logger.log(Level.INFO, "initialized Process manager \n");
    }

    /**
     * Shutdown process manager.
     */
    public void shutdown() {

        lock.lock();

        Long initPID = Long.valueOf(1);
        InitProcess init = (InitProcess) runnableMap.get(initPID);
        if (init != null)
            init.setShutdown(true);

        Set<Long> keySet = runnableMap.keySet();
        List<Long> pids = new ArrayList<Long>(keySet);
        pids.remove(initPID);

        Collections.sort(pids, new Comparator<Long>() {

            public int compare(Long o1, Long o2) {
                return -1 * o1.compareTo(o2);
            }

        });

        for (Long pid : pids) {
            Thread thread = threadMap.get(pid);
            IProcess process = runnableMap.get(pid);

            thread.interrupt();
            process.getOutput().shutdown();
            process.setShutdown(true);
            process.close();
        }
        lock.unlock();
    }

    /**
     * Shedule foreground process.
     * 
     */
    public Long shedule(IProcess process, IProcess parent) {

        lock.lock();

        Thread t = new Thread(process);
        Long pid = PIDGenerator++;
        process.setPid(pid);
        process.setParent(parent.getPid());
        parent.addChild(pid);
        Long sessionId = process.getSession().getId();
        threadMap.put(pid, t);
        runnableMap.put(pid, process);
        sessionPidMap.put(sessionId, pid);

        lock.unlock();

        t.start();

        while (!parent.isShutdown()) {
            List<Long> childs = parent.getChilds();
            if (childs.size() > 0) {

                // if thread is alive try join on another foreground child.
                Long key = null;
                List<Long> removedChild = new ArrayList<Long>();
                for (Long child : childs) {
                    if (runnableMap.get(child) == null) {
                        removedChild.add(child);
                    } else if (runnableMap.get(child).isBackground())
                        continue;
                    else
                        key = child;
                }
                parent.getChilds().removeAll(removedChild);
                Thread thread = threadMap.get(key);
                if (thread != null) {
                    try {
                        // wait until childs is running
                        thread.join();

                        // child end - remove child and try wait for another
                        lock.lock();
                        IProcess actualParent = runnableMap.get(process.getParent());
                        if(actualParent != null){
                            actualParent.getChilds().remove(pid);
                        }
                        runnableMap.remove(pid);
                        threadMap.remove(pid);
                        sessionPidMap.remove(sessionId);
                        lock.unlock();

                    } catch (InterruptedException e) {

                        // sleeping parent is interupted from wait
                        lock.lock();
                        // childs get parent of parent, parent is ended
                        IProcess parentOfParent = runnableMap.get(parent.getParent());
                        List<Long> childsOfProcess = parent.getChilds();
                        if (parentOfParent != null) {

                            for (Long child : childsOfProcess) {
                                IProcess childProc = runnableMap.get(child);
                                if (childProc != null) {
                                    childProc.setParent(parentOfParent.getPid());
                                    parentOfParent.getChilds().add(child);
                                }
                            }
                            parentOfParent.getChilds().remove(parent.getPid());
                        }

                        runnableMap.remove(parent.getPid());
                        sessionPidMap.remove(parent.getSession().getId());
                        parent.setShutdown(true);

                        lock.unlock();
                        return pid;
                    }
                } else {
                    break; // no foreground process in childs
                }
            } else {
                break; // no process in childs
            }

        }
        return pid;
    }

    /**
     * Shedule background process.
     * 
     */
    public void sheduleBackgroud(IProcess process, IProcess parent) {

        lock.lock();

        Thread t = new Thread(process);
        Long pid = PIDGenerator++;
        process.setPid(pid);
        process.setParent(parent.getPid());
        process.setBackground(true);
        parent.addChild(pid);
        threadMap.put(pid, t);
        runnableMap.put(pid, process);

        lock.unlock();

        t.start();

    }

    /**
     * Method for ending background process and remove it from process table.
     * 
     */
    public void endBackground(Long pid) {

        lock.lock();

        IProcess process = runnableMap.remove(pid);
        threadMap.remove(pid);

        if (process != null) {
            sessionPidMap.remove(process.getSession().getId());
            Long parent = process.getParent();
            IProcess parentProcess = runnableMap.get(parent);

            if (parentProcess != null) {
                parentProcess.getChilds().remove(pid);
            }
        }

        lock.unlock();

    }

    /**
     * Method for register init process when system is started.
     * 
     */
    protected void addThread(Thread t, IProcess process) {

        lock.lock();

        Long pid = PIDGenerator++;
        process.setPid(pid);
        process.setParent(null);
        threadMap.put(pid, t);
        runnableMap.put(pid, process);
        sessionPidMap.put(process.getSession().getId(), pid);

        lock.unlock();

    }

    /**
     * Kill process with pid.
     * 
     */
    public void kill(Long pid) {

        // find killed thread and remove from map
        IProcess proc = runnableMap.get(pid);

        if (proc != null) {

            if (proc.getPipedProcesses() == null) {
                IProcess process = runnableMap.remove(pid);
                Thread thread = threadMap.remove(pid);
                sessionPidMap.remove(proc.getSession().getId());
                // find parent
                IProcess parent = runnableMap.get(proc.getParent());

                if (parent == null) {
                    thread.interrupt();
                    process.close();
                    return;
                }
                // if process exist when u try killing him - get his child and get it his parent
                List<Long> childs = process.getChilds();
                for (Long child : childs) {
                    IProcess iProcess = runnableMap.get(child);
                    if (iProcess != null)
                        iProcess.setParent(parent.getPid());
                }

                parent.getChilds().remove(process.getPid());
                parent.getChilds().addAll(childs);
                process.setOutput(new EmptyOutputService());
                thread.interrupt();
                process.close();
            } else {
                List<Process> pipedProcesses = proc.getPipedProcesses();
                Collections.sort(pipedProcesses);
                for (Process process : pipedProcesses) {
                    process.setPipedProcesses(null);
                    process.setOutput(new EmptyOutputService());
                    kill(process.getPid());
                }
            }
        }

    }

    /**
     * Create map of weak copy of proceses only for PS command.
     * 
     */
    public Map<Long, IProcess> getRunnableMap() {

        lock.lock();

        Map<Long, IProcess> map = new HashMap<Long, IProcess>();
        for (Entry<Long, IProcess> entry : runnableMap.entrySet())
            map.put(entry.getKey(), entry.getValue().getWeakCopy());

        lock.unlock();

        return map;
    }

    /**
     * Create map of process states only for PS command.
     * 
     */
    public Map<Long, String> getThreadStatusMap() {

        lock.lock();

        Map<Long, String> map = new HashMap<Long, String>();
        for (Entry<Long, Thread> entry : threadMap.entrySet())
            map.put(entry.getKey(), entry.getValue().getState().name());

        lock.unlock();

        return map;
    }

    /**
     * Signal for interrupt actual foreground process.
     * 
     */
    public void signalInterupt(Session session) {

        Long pid = sessionPidMap.get(session.getId());
        if (pid == null)
            return;

        lock.lock();
        IProcess process = runnableMap.get(pid);
        lock.unlock();
        if (!process.getCommand().equalsIgnoreCase(new ShellProcess().getCommand()))
            kill(pid);

    }

}
