package os.process.shell;

import java.util.List;

import os.io.OutputService;
import os.process.IProcess;
import os.process.Process;
import os.process.ProcessManager;
import os.process.shell.login.LoginProcess;
import os.session.SessionManager;

/**
 * Shell process.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 */
public class ShellProcess extends Process {

    private static final String COMMAND = "SHELL";

    private SyntaxAnalyzator syntax;

    public ShellProcess() {

        setSession(SessionManager.get().getNewSession());

        this.syntax = new SyntaxAnalyzator();
    }

    public void run() {

        if (getOutput().getType() != OutputService.PIPE) {

            this.syntax.setSession(getSession());
            this.syntax.setInput(getInput());
            this.syntax.setOutput(getOutput());
            this.syntax.setError(getError());

            preprocessArgumentsAndParameters();

            // read command from input until thread is interruped or system is in shutdow.
            while (!Thread.currentThread().isInterrupted() && !isShutdown()) {
                // if session is not logged - force login
                if (!getSession().isLogged()) {
                    runLoginProcess();
                    setLogged(true);
                    continue;
                }

                // write prepromt string "login~actualDir"
                getOutput().write(getSession().getPromtString() + "\n");

                // read input from standard input
                String command = getInput().readLine();

                // if command is null its interupt
                if (command == null) {
                    break;
                }

                command = command.trim();

                if (command.equalsIgnoreCase("exit")) {
                    getSession().getHistory().addHistoryItem(getSession().getDirectory(), command);
                    getOutput().write(command + "\n");
                    close();
                    break;

                    // analyze command
                } else {

                    getSession().getHistory().addHistoryItem(getSession().getDirectory(), command);
                    getOutput().write(command + "\n");

                    // syntax analyzator get list of process
                    List<Process> listOfProcesses = syntax.analyze(command);

                    if (listOfProcesses != null) {
                        for (Process process : listOfProcesses) {

                            if (process instanceof LoginProcess)
                                setLogged(true);

                            if (process.isBackground()) {
                                ProcessManager.get().sheduleBackgroud(process, this);
                            } else {
                                ProcessManager.get().shedule(process, this);
                            }
                        }
                    }
                }
            }

            if (isLogged())
                SessionManager.get().logoutSession(getSession());
            
            Thread.currentThread().interrupt();
            close();
        }
    }

    private void preprocessArgumentsAndParameters() {

        // shell isn't accept arguments
        if (getArgs() != null && getArgs().size() > 0) {
            getError().write("shell: invalid option, use 'help shell'\n");
            Thread.currentThread().interrupt();
            return;
        }

        // shell accept only parameter for login ( -l).
        for (String parameter : getParams()) {
            // login via parameter
            if (parameter.equals("-l")) {
                runLoginProcess();
                setLogged(true);
            } else {
                getError().write("shell: " + parameter + ": invalid option, use 'help shell'\n");
                Thread.currentThread().interrupt();
                return;
            }
        }
    }

    /**
     * Prepare and run login process.
     */
    private void runLoginProcess() {

        LoginProcess login = new LoginProcess();
        login.setInput(getInput());
        login.setOutput(getOutput());
        login.setError(getError());
        login.setSession(getSession());

        ProcessManager.get().shedule(login, this);
    }

    public String getCommand() {
        return COMMAND;
    }

    public IProcess getClone() {
        return new ShellProcess();
    }

    public IProcess getWeakCopy() {

        ShellProcess clone = new ShellProcess();
        clone.setArguments(getArgs());
        clone.setParameters(getParams());
        clone.setPid(getPid());
        clone.setParent(getParent());
        clone.setChilds(getChilds());
        return clone;
    }

}
