package os.process.shell;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import os.file.FileManager;
import os.io.FileInputService;
import os.io.FileOutputService;
import os.io.InputService;
import os.io.OutputService;
import os.io.PipeInputOutputService;
import os.process.Process;
import os.process.shell.cd.CdProcess;
import os.process.shell.login.LoginProcess;
import os.session.Session;

/**
 * 
 * Syntax analyzator for command in shell.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 * 
 */
public class SyntaxAnalyzator {

    private Session session;

    // standard input, output and error streams
    private InputService input;
    private OutputService output;
    private OutputService error;

    private String[] parsedLine;
    private String token;
    private int parsedLineLength;
    private int index;
    private boolean isSyntaxError;
    private boolean isFileError;
    private Process process;
    private boolean hasFirstProcessOutputToFile;

    public SyntaxAnalyzator() {
        parsedLine = null;
        token = null;
        parsedLineLength = 0;
        index = 0;
        isSyntaxError = false;
        process = null;
        hasFirstProcessOutputToFile = false;
    }

    public void setSession(Session session) {
        this.session = session;
    }

    public void setInput(InputService input) {
        this.input = input;
    }

    public void setOutput(OutputService output) {
        this.output = output;
    }

    public void setError(OutputService error) {
        this.error = error;
    }

    /**
     * Syntax analyze of given line, return list of processes. <br/>
     * 
     * First, line is analyzed by lexical analizator, that add spaces to line before and behind each operand. <br/>
     * Then the line is split by spaces to tokens. These tokens are analyzed to: <br/>
     * programs name, parameters, argumets and operands.<br/>
     * 
     */
    public List<Process> analyze(String line) {

        List<Process> listOfProcesses = new ArrayList<Process>();

        parsedLine = lexical(line);
        parsedLineLength = parsedLine.length;
        index = 0;
        isSyntaxError = false;
        isFileError = false;

        if (line.equalsIgnoreCase(""))
            return null;

        // finite-state machine of syntax analyze
        if (isProcess()) {

            if (!process.getCommand().equalsIgnoreCase(new LoginProcess().getCommand()) && !process.getCommand().equalsIgnoreCase(new CdProcess().getCommand())) {
                process.setSession(session.cloneWithUniqueId()); // other proceses get copy of session with unique session id.
            } else {
                process.setSession(session); // login use origin session for login action
            }

            process.setInput(input);
            process.setOutput(output);
            process.setError(error);

            listOfProcesses.add(process);

            while (isParameter()) {
                process.addParameter(token);
            }
            while (isArgument()) {
                process.addArgument(token);
            }

            while (isInputOver()) {

                if (isArgument()) {

                    try {
                        FileInputService fis = FileManager.get().openFileAsInput(session.getDirectory(), token);

                        if (fis != null) {
                            if (!process.getCommand().equals("SHELL")) {
                                process.setInput(fis);
                            }
                        } else {
                            error.write("OSMRR: " + token + ": No such file or directory\n");
                            isFileError = true;
                            break;
                        }
                    } catch (FileNotFoundException e) {
                        error.write("OSMRR: " + token + ": No such file or directory\n");
                        isFileError = true;
                        break;
                    }

                } else {
                    isSyntaxError = true;
                    break;
                }
            }

            while (isOutput()) {

                if (isOutputOver()) {

                    if (isArgument()) {

                        try {
                            FileOutputService fos = FileManager.get().openFileAsOutput(session.getDirectory(), token, false, false);

                            if (fos != null) {
                                if (!process.getCommand().equals("SHELL")) {
                                    process.setOutput(fos);
                                    hasFirstProcessOutputToFile = true;
                                }
                            } else {
                                error.write("OSMRR: " + token + ": Can't open file for writing\n");
                                isFileError = true;
                                break;
                            }
                        } catch (IOException e) {
                            error.write("OSMRR: " + token + ": Can't open file for writing\n");
                            isFileError = true;
                            break;
                        }
                    } else {
                        isSyntaxError = true;
                        break;
                    }
                }

                if (isOutputToEndOver()) {

                    if (isArgument()) {

                        try {
                            FileOutputService fos = FileManager.get().openFileAsOutput(session.getDirectory(), token, false, true);

                            if (fos != null) {
                                if (!process.getCommand().equals("SHELL")) {
                                    process.setOutput(fos);
                                    hasFirstProcessOutputToFile = true;
                                }
                            } else {
                                error.write("OSMRR: " + token + ": Can't open file for writing\n");
                                isFileError = true;
                                break;
                            }

                        } catch (IOException e) {
                            error.write("OSMRR: " + token + ": Can't open file for writing\n");
                            isFileError = true;
                            break;
                        }

                    } else {
                        isSyntaxError = true;
                        break;
                    }
                }

                if (isErrorOutputOver()) {

                    if (isArgument()) {

                        output.write("OSMRR: redirection of error output to file is disabled\r\n");

                        // turn off redirecton of error output
                        if (false) {
                            try {
                                FileOutputService fos = FileManager.get().openFileAsOutput(session.getDirectory(), token, true, false);

                                if (fos != null) {
                                    if (!process.getCommand().equals("SHELL")) {
                                        process.setError(fos);
                                    }
                                } else {
                                    error.write("OSMRR: " + token + ": Can't open file for writing\n");
                                    isFileError = true;
                                    break;
                                }
                            } catch (IOException e) {
                                error.write("OSMRR: " + token + ": Can't open file for writing\n");
                                isFileError = true;
                                break;
                            }
                        }

                    } else {
                        isSyntaxError = true;
                        break;
                    }
                }
            }

            while (isPiped()) {

                if (!hasFirstProcessOutputToFile) {

                    PipeInputOutputService pipe = new PipeInputOutputService();
                    process.setOutput(pipe);
                    process.setBackground(true);

                    if (isProcess()) {

                        process.setSession(session.cloneWithUniqueId());

                        process.setInput(pipe);
                        process.setOutput(output);
                        process.setError(error);

                        listOfProcesses.add(process);

                        while (isParameter()) {
                            process.addParameter(token);
                        }

                        while (isArgument()) {
                            process.addArgument(token);
                        }

                    } else {
                        error.write("OSMRR: " + token + ": command not found\n");

                        return null;
                    }

                } else {

                    if (isProcess()) {

                        // if first process has output to file and is pipped - skip other process in pipe
                        while (isParameter()) {
                        }

                        while (isArgument()) {
                        }

                    } else {
                        error.write("OSMRR: " + token + ": command not found\n");

                        return null;
                    }
                }
            }

            while (isOutput()) {

                if (isOutputOver()) {

                    if (isArgument()) {

                        try {
                            FileOutputService fos = FileManager.get().openFileAsOutput(session.getDirectory(), token, false, false);

                            if (fos != null) {
                                process.setOutput(fos);
                            } else {
                                error.write("OSMRR: " + token + ": Can't open file for writing\n");
                                isFileError = true;
                                break;
                            }
                        } catch (IOException e) {
                            error.write("OSMRR: " + token + ": Can't open file for writing\n");
                            isFileError = true;
                            break;
                        }

                    } else {
                        isSyntaxError = true;
                        break;
                    }
                }

                if (isOutputToEndOver()) {

                    if (isArgument()) {

                        try {
                            FileOutputService fos = FileManager.get().openFileAsOutput(session.getDirectory(), token, false, true);

                            if (fos != null) {
                                process.setOutput(fos);
                            } else {
                                error.write("OSMRR: " + token + ": Can't open file for writing\n");
                                isFileError = true;
                                break;
                            }
                        } catch (IOException e) {
                            error.write("OSMRR: " + token + ": Can't open file for writing\n");
                            isFileError = true;
                            break;
                        }

                    } else {
                        isSyntaxError = true;
                        break;
                    }
                }

                if (isErrorOutputOver()) {

                    if (isArgument()) {

                        output.write("OSMRR: redirection of error output to file is disabled\r\n");

                        // turn off redirecton of error output
                        if (false) {
                            try {
                                FileOutputService fos = FileManager.get().openFileAsOutput(session.getDirectory(), token, true, false);

                                if (fos != null) {
                                    process.setError(fos);
                                } else {
                                    error.write("OSMRR: " + token + ": Can't open file for writing\n");
                                    isFileError = true;
                                    break;
                                }
                            } catch (IOException e) {
                                error.write("OSMRR: " + token + ": Can't open file for writing\n");
                                isFileError = true;
                                break;
                            }
                        }

                    } else {
                        isSyntaxError = true;
                        break;
                    }
                }
            }

            if (isAsynchroneous()) {
                for (Process process : listOfProcesses)
                    process.setBackground(true);
            }

            if (isFileError)
                return null;

            // if there is syntax error of one or more tokens lef, there is error in input line
            if (hasNext() || isSyntaxError) {

                if (index == parsedLineLength && isSyntaxError) {
                    token = "newline";
                }

                error.write("OSMRR: syntax error near unexpected token '" + token + "'\n");

                return null;
            }

            // if there is more than one processes (piped), set piped processes to all processes
            if (listOfProcesses.size() > 1) {
                for (Process proc : listOfProcesses) {
                    proc.setPipedProcesses(listOfProcesses);
                }
            }

            return listOfProcesses;

        } else {
            error.write("OSMRR: " + token + ": command not found\n");

            return null;
        }
    }

    /**
     * Lexical analyze of input line.
     * 
     * Line is analyzed by lexical analizator, that add spaces to line before and behind each operand. <br/>
     * Then the line is split by spaces to tokens and returned to syntax analyzator.
     * 
     */
    private String[] lexical(String line) {

        StringBuilder sb = new StringBuilder();

        // add spaces before and behind all operands
        for (int c = 0; c < line.length(); c++) {

            if (line.charAt(c) == '<' || line.charAt(c) == '|' || line.charAt(c) == '>' || line.charAt(c) == '&') {

                if (c > 0 && line.charAt(c - 1) != ' ') {
                    sb.append(' ');
                }

                sb.append(line.charAt(c));

                if (c < (line.length() - 1) && line.charAt(c + 1) == '>') {
                    sb.append('>');
                    c++;
                }

                if (c < (line.length() - 1) && line.charAt(c + 1) != ' ') {
                    sb.append(' ');
                }

            } else if (line.charAt(c) == '2' && c < (line.length() - 1) && line.charAt(c + 1) == '>') {

                if (c > 0 && line.charAt(c - 1) != ' ') {
                    sb.append("2 >");
                } else {
                    sb.append("2>");
                }

                c++;
                if (c < (line.length() - 1) && line.charAt(c + 1) != ' ') {
                    sb.append(' ');
                }

            } else {
                sb.append(line.charAt(c));
            }
        }

        // split input line by spaces
        List<String> tokens = new ArrayList<String>();
        int begin = 0;
        boolean isQuotationMark = false; // uvozovky "

        line = sb.toString();
        for (int c = 0; c < line.length(); c++) {

            if (line.charAt(c) == '"') {
                isQuotationMark = !isQuotationMark;

                if (c < line.length() - 1) {
                    continue;
                }
            }

            if (line.charAt(c) == ' ' && !isQuotationMark) {
                tokens.add(line.substring(begin, c));

                begin = c + 1;

                continue;
            }

            if (c == line.length() - 1) {
                tokens.add(line.substring(begin, c + 1));
            }
        }

        return tokens.toArray(new String[tokens.size()]);
    }

    /**
     * Test if there is more tokens.
     * 
     */
    private boolean hasNext() {

        if (!isSyntaxError && !isFileError && index < parsedLineLength) {
            token = parsedLine[index];

            return true;
        }

        return false;
    }

    /**
     * Test if token is process command.
     * 
     */
    private boolean isProcess() {
        process = null;

        if (!hasNext())
            return false;

        for (Process runableProcess : Process.runnable) {

            if (runableProcess.getCommand() != null && token.equalsIgnoreCase(runableProcess.getCommand())) {
                process = (Process) runableProcess.getClone();

                if (process.getCommand().equals("KILL")) {
                    process.setBackground(true);
                }

                index++;
                return true;
            }
        }

        return false;
    }

    /**
     * Test if token is parameter.
     * 
     */
    private boolean isParameter() {

        if (!hasNext()) {
            return false;
        }

        if (token.startsWith("-")) {
            index++;
            return true;
        }

        return false;
    }

    /**
     * Test if token is argument.
     * 
     */
    private boolean isArgument() {

        if (!hasNext()) {
            return false;
        }

        if (token.equalsIgnoreCase("<") || token.equalsIgnoreCase("|") || token.equalsIgnoreCase(">") || token.equalsIgnoreCase(">>") || token.equalsIgnoreCase("2>") || token.equalsIgnoreCase("&")) {
            return false;
        }

        index++;
        return true;
    }

    /**
     * Test if token is input from file.
     * 
     */
    private boolean isInputOver() {

        if (!hasNext()) {
            return false;
        }

        if (token.equalsIgnoreCase("<")) {
            index++;
            return true;
        }

        return false;
    }

    /**
     * Test if token is pipe.
     * 
     */
    private boolean isPiped() {

        if (!hasNext()) {
            return false;
        }

        if (token.equalsIgnoreCase("|")) {
            index++;
            return true;
        }

        return false;
    }

    /**
     * Test if token is output to file, output to end of file or error output to file.
     * 
     */
    private boolean isOutput() {

        if (!hasNext()) {
            return false;
        }

        if (token.equalsIgnoreCase(">") || token.equalsIgnoreCase(">>") || token.equalsIgnoreCase("2>")) {
            return true;
        }

        return false;
    }

    /**
     * Test if token is output to file.
     * 
     */
    private boolean isOutputOver() {

        if (!hasNext()) {
            return false;
        }

        if (token.equalsIgnoreCase(">")) {
            index++;
            return true;
        }

        return false;
    }

    /**
     * Test if token is output to end of file.
     * 
     */
    private boolean isOutputToEndOver() {

        if (!hasNext()) {
            return false;
        }

        if (token.equalsIgnoreCase(">>")) {
            index++;
            return true;
        }

        return false;
    }

    /**
     * Test if token is error output to file.
     * 
     */
    private boolean isErrorOutputOver() {

        if (!hasNext()) {
            return false;
        }

        if (token.equalsIgnoreCase("2>")) {
            index++;
            return true;
        }

        return false;
    }

    /**
     * Test if token is &, background processes.
     * 
     */
    private boolean isAsynchroneous() {
        if (!hasNext())
            return false;

        if (token.equalsIgnoreCase("&")) {
            index++;
            return true;
        }

        return false;
    }
}
