package os.process.shell.cat;

import java.io.FileNotFoundException;

import os.file.FileManager;
import os.io.FileInputService;
import os.io.InputService;
import os.io.OutputService;
import os.process.IProcess;
import os.process.Process;

/**
 * Cat proces for reading data from file.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 * 
 */
public class CatProcess extends Process {

    private static final String COMMAND = "CAT";

    private static final String LINENUMBERFORMAT = "%-6d";
    private boolean addNumbersNonEmptyLines = false; // -b
    private boolean showEnds = false; // -E
    private boolean addNumbersAllLines = false; // -n
    private boolean squezeEmptyLines = false; // -s
    private boolean wasEmptyLine = false;
    private int lineNumber;

    public CatProcess() {
        lineNumber = 1;
    }

    public void run() {

        String inputFile = null;

        // set parameters
        for (String parameter : getParams()) {

            if (parameter.equals("-b")) {

                addNumbersNonEmptyLines = true;
                addNumbersAllLines = false;

            } else if (parameter.equals("-E")) {

                showEnds = true;

            } else if (parameter.equals("-n")) {

                if (!addNumbersNonEmptyLines) {
                    addNumbersAllLines = true;
                }

            } else if (parameter.equals("-s")) {

                squezeEmptyLines = true;

            } else {

                getError().write("cat: " + parameter + ": invalid option, use 'help cat'\n");
                close();
                Thread.currentThread().interrupt();
                return;
            }
        }

        // test whether program has argumets
        if (getArgs() != null && getArgs().size() > 0) {

            // go through all arguments and print content if it is file
            for (String argument : getArgs()) {
                inputFile = argument;

                if (inputFile != null) {

                    try {
                        FileInputService fis = FileManager.get().openFileAsInput(getSession().getDirectory(), inputFile);

                        if (fis != null) {
                            if (getInput() != null)
                                getInput().close();

                            setInput(fis);
                        } else {
                            getError().write("cat: " + inputFile + ": No such file or directory\r\n");
                            close();
                            Thread.currentThread().interrupt();
                            return;
                        }

                    } catch (FileNotFoundException e) {
                        getError().write("cat: " + inputFile + ": No such file or directory\r\n");
                        close();
                        Thread.currentThread().interrupt();
                        return;
                    }
                }

                readFromInputWriteToOutput();

                getInput().close();
            }
        } else {
            readFromInputWriteToOutput();
        }

        close();
    }

    
    /**
     * Read data from input stream and print it to output stream in specific format if flags are set.
     * 
     */
    private void readFromInputWriteToOutput() {

        String line;

        while ((line = getInput().readLine()) != null && !Thread.currentThread().isInterrupted()) {

            if (getInput().getType() == InputService.STANDARD && getOutput().getType() == OutputService.STANDARD) {
                getOutput().write(line + "\r\n");
            }

            if (!squezeEmptyLines || squezeEmptyLines && (!wasEmptyLine || wasEmptyLine && !line.trim().equals(""))) {

                wasEmptyLine = line.trim().equals("");

                if ((addNumbersNonEmptyLines && !line.trim().equals("")) || addNumbersAllLines) {
                    getOutput().write(String.format(LINENUMBERFORMAT, lineNumber++));
                }
                getOutput().write(line);

                if (showEnds) {
                    getOutput().write("$");
                }
                getOutput().write("\r\n");
            }
        }
    }

    public String getCommand() {
        return COMMAND;
    }

    public IProcess getClone() {
        return new CatProcess();
    }

    public IProcess getWeakCopy() {
        CatProcess clone = new CatProcess();
        clone.setArguments(getArgs());
        clone.setParameters(getParams());
        clone.setPid(getPid());
        clone.setParent(getParent());
        clone.setChilds(getChilds());
        return clone;
    }
}
