package os.process.shell.cp;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;

import os.file.FileManager;
import os.io.FileInputService;
import os.io.FileOutputService;
import os.io.OutputService;
import os.process.IProcess;
import os.process.Process;

/**
 * Basic CP program - copy one file to another file via bytes.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 * 
 */
public class CopyProcess extends Process {

    private static final String ACTUAL_DIR_PREFIX = "./";
    private static final String COMMAND = "CP";

    private static Logger logger = Logger.getLogger(CopyProcess.class.getName());

    private boolean copyWithContent = false;

    public CopyProcess() {
    }

    public void run() {

        if (getOutput().getType() != OutputService.PIPE) {

            if (getArgs().size() > 2) {
                getError().write("cp: wrong number of arguments, use 'help cp'\n");
                close();
                Thread.currentThread().interrupt();
                return;
            }

            if (getArgs().size() < 2) {
                getError().write("cp: missing argument/s, use 'help cp'\n");
                close();
                Thread.currentThread().interrupt();
                return;
            }

            // get source file path from arguments
            String source = getArgs().get(0);
            if (source.startsWith(ACTUAL_DIR_PREFIX)) {
                source = source.substring(2);
            }

            // get target file path from arguments
            String target = getArgs().get(1);
            if (target.startsWith(ACTUAL_DIR_PREFIX)) {
                target = target.substring(2);
            }

            String path = getSession().getDirectory();

            // prepare source path for use
            if (source.equalsIgnoreCase(".")) {
                // if path is actual dir get actual dir from session
                source = getSession().getDirectory();
            } else {
                // else create final source path
                String file = FileManager.get().getPathBasedOnOS(source);
                if (file == null) {
                    if (!path.endsWith(File.separator)) {
                        source = FileManager.get().getPathBasedOnOS(path + File.separator + source);
                    } else {
                        source = FileManager.get().getPathBasedOnOS(path + source);
                    }
                } else {
                    source = file;
                }
            }

            // prepare target path for user
            if (target.equalsIgnoreCase(".")) {
                // if path is actual dir get actual dir from session
                target = path;
            } else {
                // else create final source path
                String file = FileManager.get().getPathBasedOnOS(target);
                if (file == null) {
                    if (!path.endsWith(File.separator)) {
                        target = FileManager.get().getPathBasedOnOS(path + File.separator + target);
                    } else {
                        target = FileManager.get().getPathBasedOnOS(path + target);
                    }
                } else {
                    target = file;
                }
            }

            // test if source is exists.
            if (!FileManager.get().isPath(source)) {
                getError().write("cp: wrong source file or dir\n");
                close();
                Thread.currentThread().interrupt();
                return;
            }

            File sourceFile = FileManager.get().getFileOrDir(source);
            File targetFile;
            if (sourceFile.isDirectory())
                targetFile = FileManager.get().getOrCreateDir(target);
            else
                targetFile = FileManager.get().getOrCreateFile(target);

            for (String parameter : getParams()) {
                if (parameter.equals("-r")) {
                    copyWithContent = true;
                } else {
                    getError().write("cp: " + parameter + ": invalid option, use 'help cp'\n");
                    close();
                    Thread.currentThread().interrupt();
                    return;
                }
            }

            if (copyWithContent) {
                File targetDir = FileManager.get().getOrCreateDir(targetFile, sourceFile.getName());
                targetDir.mkdir();
                copy(sourceFile, targetDir);
            } else {
                copy(sourceFile, targetFile);
            }

            close();
        }
    }
    
    /**
     * Copy from source to target file or dir.
     */
    private void copy(File source, File target) {

        if (getOutput().getType() != OutputService.PIPE) {

            // if source is directory
            if (source.isDirectory()) {

                // test if source and target is same.
                if (source.getAbsolutePath().equalsIgnoreCase(target.getAbsolutePath())) {
                    getError().write("cp: cannot copy file to itself\n");
                    logger.info("Cannot copy file to itself\n");
                    return;
                }

                // if target directory isn't exist create it.
                if (!target.exists()) {
                    target.mkdir();
                }

                String[] filesInDir = source.list();

                // copy recursively files from source dir to target dir
                for (String file : filesInDir) {

                    if (Thread.currentThread().isInterrupted())
                        break;

                    File sourceFile = FileManager.get().getFileOrDir(source, file);

                    File targetFile;
                    if (sourceFile.isDirectory())
                        targetFile = FileManager.get().getOrCreateDir(target, file);
                    else
                        targetFile = FileManager.get().getOrCreateFile(target, file);

                    copy(sourceFile, targetFile);
                }

            } else {
                // copy just file to file or directory
                copyFileToDest(source, target);
            }
        }
    }
    
    private void copyFileToDest(File source, File target) {

        // get or create target file or directory
        if (target.isDirectory()) {
            target = FileManager.get().getOrCreateFile(target + File.separator + source.getName());
        }

        try {
            // prepare input and output
            setInput(new FileInputService(source));
            setOutput(new FileOutputService(target, false, false));
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }

        // copy source to target.
        int bufferSize = 1024;
        byte[] bytes = new byte[bufferSize];
        int c = getInput().read(bytes);
        while (c != -1 && !Thread.currentThread().isInterrupted()) {

            if (c < bufferSize) {
                bytes = Arrays.copyOf(bytes, c);
            }

            getOutput().write(bytes);
            c = getInput().read(bytes);
        }
        getInput().close();
        getOutput().close();
    }

    public String getCommand() {
        return COMMAND;
    }

    public IProcess getClone() {
        return new CopyProcess();
    }

    public IProcess getWeakCopy() {

        CopyProcess clone = new CopyProcess();
        clone.setArguments(getArgs());
        clone.setParameters(getParams());
        clone.setPid(getPid());
        clone.setParent(getParent());
        clone.setChilds(getChilds());
        return clone;
    }

}
