package os.process.shell.ls;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.Date;

import os.file.FileManager;
import os.process.IProcess;
import os.process.Process;

/**
 * Ls process for list directory.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 * 
 */
public class LsProcess extends Process {

    // formats of printed lines
    private static final String LINE_END = "\r\n";
    private static final String SHORTOUTPUTLINE = "%-25s   %-25s   %-25s";
    private static final String LONGOUTPUTLINE = "%-3s %-11d %-10s %-20s";
    private static final String COMMAND = "LS";

    private static final SimpleDateFormat dateFormat = new SimpleDateFormat("MM/dd/yyyy");

    private boolean isAll = false;
    private boolean isLongPrint = false;

    public LsProcess() {
    }

    public void run() {

        // go through parameters and set flags
        for (String parameter : getParams()) {

            if (parameter.equals("-a")) {
                isAll = true;

                continue;
            } else if (parameter.equals("-l")) {
                isLongPrint = true;

                continue;
            } else {
                getError().write("ls: " + parameter + ": invalid option, use 'help ls'\n");
                close();
                Thread.currentThread().interrupt();
                return;
            }
        }

        String path = getSession().getDirectory();

        // test if there are arguments and if last argument is a path
        if (getArgs() != null && getArgs().size() > 0) {
            String actualDirectory = getSession().getDirectory();
            String newDirectory = getArgs().get(getArgs().size() - 1).replace("\"", "");

            if (!FileManager.get().isAbsoluteDir(newDirectory)) {
                if (!actualDirectory.endsWith(File.separator)) {
                    newDirectory = actualDirectory + File.separator + newDirectory;
                } else {
                    newDirectory = actualDirectory + newDirectory;
                }
            }

            newDirectory = FileManager.get().getPathBasedOnOS(newDirectory);

            if (newDirectory != null && FileManager.get().isDir(newDirectory)) {
                path = newDirectory;
            } else {
                getError().write("ls: " + newDirectory + ": no such file or directory\n");
                close();
                Thread.currentThread().interrupt();
                return;
            }
        }

        // get list of files/folders in path given as argument or from actual directory
        File[] listOfFiles = FileManager.get().getListOfFiles(path);
        String fileName;
        int count = 0;
        String[] fileRow = { "", "", "" };

        // print all files and folders in specific format if flags are set
        if (listOfFiles != null) {

            for (int i = 0; i < listOfFiles.length; i++) {

                if (Thread.currentThread().isInterrupted())
                    break;

                File file = listOfFiles[i];
                fileName = file.getName();

                if (isAll || !file.isHidden()) {

                    if (isLongPrint) {

                        // file atributts - directory/executible/- , read/-, write/-
                        StringBuilder fileAtributes = new StringBuilder();

                        if (file.isDirectory()) {
                            fileAtributes.append('d');
                        } else {
                            if (file.canExecute()) {
                                fileAtributes.append('e');
                            } else {
                                fileAtributes.append('-');
                            }
                        }

                        if (file.canRead()) {
                            fileAtributes.append('r');
                        } else {
                            fileAtributes.append('-');
                        }

                        if (file.canWrite()) {
                            fileAtributes.append('w');
                        } else {
                            fileAtributes.append('-');
                        }

                        // file size
                        Long size = file.length();

                        // date of last file modification
                        Date dateLastModified = new Date(file.lastModified());
                        String lastModified = dateFormat.format(dateLastModified);

                        String string = String.format(LONGOUTPUTLINE, fileAtributes.toString(), size, lastModified, fileName);
                        getOutput().write(string + LINE_END);
                    } else {

                        // short print
                        fileRow[count++] = fileName;

                        if (count == 3 || i == listOfFiles.length - 1) {
                            count = 0;
                            getOutput().write(String.format(SHORTOUTPUTLINE, fileRow[0], fileRow[1], fileRow[2]) + LINE_END);
                            fileRow[0] = "";
                            fileRow[1] = "";
                            fileRow[2] = "";
                        }
                    }
                }
            }

            getOutput().write("\n");
        }

        close();
    }

    public String getCommand() {
        return COMMAND;
    }

    public IProcess getClone() {
        return new LsProcess();
    }

    public IProcess getWeakCopy() {

        LsProcess clone = new LsProcess();
        clone.setArguments(getArgs());
        clone.setParameters(getParams());
        clone.setPid(getPid());
        clone.setParent(getParent());
        clone.setChilds(getChilds());
        return clone;

    }

}
