package os.process.shell.sort;

import java.io.FileNotFoundException;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Locale;

import os.file.FileManager;
import os.io.FileInputService;
import os.io.InputService;
import os.io.OutputService;
import os.process.IProcess;
import os.process.Process;

/**
 * Sort process for sorting data.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 * 
 */
public class SortProcess extends Process {

    private static final String COMMAND = "SORT";
    private boolean stringTrim = false;
    private boolean revert = false;
    private boolean random = false;

    public SortProcess() {
    }

    public void run() {

        try {
            ArrayList<Line> seznam = new ArrayList<Line>();
            String inputFile = null;
            ArrayList<Comparator<Line>> compare = new ArrayList<Comparator<Line>>();

            // go through parameters, set flags
            for (String parameter : getParams()) {

                if (parameter.equals("-b")) {
                    stringTrim = true;
                } else if (parameter.equals("-r")) {
                    revert = true;
                } else if (parameter.equals("-R")) {
                    random = true;
                } else if (parameter.equals("-d")) {
                    compare.add(new AlphanumericComporator());
                } else if (parameter.equals("-g")) {
                    compare.add(new ASCIIValueComparator());
                } else if (parameter.equals("-f")) {
                    compare.add(new IgnoreCaseComparator());
                } else {
                    getError().write("sort: " + parameter + ": invalid option, use 'help sort'\n");
                    close();
                    Thread.currentThread().interrupt();
                    return;
                }
            }

            // if there are argumets and these arguments are files, take them as input for sort
            if (getArgs() != null && getArgs().size() > 0) {

                for (String args : getArgs()) {
                    inputFile = args;

                    if (Thread.currentThread().isInterrupted())
                        break;

                    if (inputFile != null) {

                        try {
                            FileInputService fis = FileManager.get().openFileAsInput(getSession().getDirectory(), inputFile);

                            if (fis != null) {
                                if (getInput() != null)
                                    getInput().close();

                                setInput(fis);
                            } else {
                                getError().write("sort: " + inputFile + ": No such file\n");
                                close();
                                Thread.currentThread().interrupt();
                                return;
                            }

                        } catch (FileNotFoundException e) {
                            getError().write("sort: " + inputFile + ": No such file\n");
                            close();
                            Thread.currentThread().interrupt();
                            return;
                        }
                    }

                    readFromInput(seznam);

                    getInput().close();

                }
            } else {
                
                // read from standard input
                readFromInput(seznam);
            }

            // sort readed data and print result to output
            if (seznam.size() > 0) {

                if (compare.size() > 0) {
                    for (Comparator<Line> komparator : compare) {

                        if (Thread.currentThread().isInterrupted())
                            break;

                        Collections.sort(seznam, komparator);
                    }
                } else {
                    Collections.sort(seznam, new CzechComparator());
                }

                if (revert) {
                    Collections.reverse(seznam);
                }

                if (random) {
                    int size = seznam.size();
                    ArrayList<Line> pomSeznam = new ArrayList<Line>();
                    for (int i = 0; i < size; i++) {

                        if (Thread.currentThread().isInterrupted())
                            break;

                        int prvek = (int) (Math.random() * seznam.size());
                        pomSeznam.add(seznam.get(prvek));
                        seznam.remove(prvek);
                    }
                    seznam = pomSeznam;
                }

                for (Line lines : seznam) {

                    if (Thread.currentThread().isInterrupted())
                        break;

                    getOutput().write(lines.getText() + "\r\n");
                }
            }

            close();
        } catch (OutOfMemoryError e) {
            getError().write(e.getMessage());
            close();
        }
    }

    /**
     * Read data for sort from standard input and save them to list.
     * 
     */
    private void readFromInput(ArrayList<Line> seznam) {

        String line;
        while ((line = getInput().readLine()) != null && !Thread.currentThread().isInterrupted()) {

            if (stringTrim) {
                seznam.add(new Line(line.trim()));
            } else {
                seznam.add(new Line(line));
            }

            if (getInput().getType() == InputService.STANDARD && getOutput().getType() == OutputService.STANDARD) {
                getOutput().write(line + "\r\n");
            }
        }
    }

    public String getCommand() {
        return COMMAND;
    }

    public IProcess getClone() {
        return new SortProcess();
    }

    public IProcess getWeakCopy() {
        SortProcess clone = new SortProcess();
        clone.setArguments(getArgs());
        clone.setParameters(getParams());
        clone.setPid(getPid());
        clone.setParent(getParent());
        clone.setChilds(getChilds());
        return clone;
    }

    /**
     * ASCII comparator for sorting.
     *
     */
    private class ASCIIValueComparator implements Comparator<Line> {

        public int compare(Line o1, Line o2) {

            int min = Math.min(o1.getText().length(), o2.getText().length());
            for (int i = 0; i < min; i++) {

                if (((int) o1.getText().charAt(i)) > ((int) o2.getText().charAt(i))) {
                    return +1;
                } else if (((int) o1.getText().charAt(i)) < ((int) o2.getText().charAt(i))) {
                    return -1;
                }

            }
            if (o1.getText().length() > o2.getText().length()) {
                return +1;
            } else if (o1.getText().length() < o2.getText().length()) {
                return -1;
            }
            return 0;
        }

    }

    /**
     * Czech comparator for sorting.
     *
     */
    private class CzechComparator implements Comparator<Line> {

        public int compare(Line o1, Line o2) {
            Collator c = Collator.getInstance(new Locale("cs", "CZ"));
            return c.compare(o1.getText(), o2.getText());
        }
    }

    /**
     * Comparator for sorting that ignores characters case.
     *
     */
    private class IgnoreCaseComparator implements Comparator<Line> {

        public int compare(Line o1, Line o2) {
            Collator c = Collator.getInstance();
            return c.compare(o1.getText().toUpperCase(new Locale("cs", "CZ")), o2.getText().toUpperCase(new Locale("cs", "CZ")));
        }
    }

    /**
     * Alphanumeric comparator for sorting.
     *
     */
    private class AlphanumericComporator implements Comparator<Line> {

        int mezera = (int) ' ';
        int tab = 9;
        int a = (int) 'a';
        int z = (int) 'z';
        int A = (int) 'A';
        int Z = (int) 'Z';
        int zeroValue = (int) '0';
        int nineValue = (int) '9';

        public int compare(Line o1, Line o2) {
            String text1 = o1.getText();
            String text2 = o2.getText();
            boolean texAlfa1 = false;
            boolean texAlfa2 = false;

            int min = Math.min(text1.length(), text2.length());
            for (int i = 0; i < min; i++) {

                texAlfa1 = isAlphanumeric(text1.charAt(i));
                texAlfa2 = isAlphanumeric(text2.charAt(i));

                char charTxt1 = cislaNaKonec((text1.charAt(i) + "").toUpperCase().charAt(0));
                char charTxt2 = cislaNaKonec((text2.charAt(i) + "").toUpperCase().charAt(0));

                if (texAlfa1 && texAlfa2) {
                    if (charTxt1 > charTxt2) {
                        return +1;
                    } else if (charTxt1 < charTxt2) {
                        return -1;
                    }
                } else if (texAlfa1) {
                    return +1;
                } else if (texAlfa2) {
                    return -1;
                } else {
                    if (charTxt1 > charTxt2) {
                        return +1;
                    } else if (charTxt1 < charTxt2) {
                        return -1;
                    }
                }
            }
            if (text1.length() > text2.length()) {
                return +1;
            } else if (text1.length() < text2.length()) {
                return -1;
            }
            return 0;

        }

        public char cislaNaKonec(char value) {
            if (zeroValue <= value && value <= nineValue) {
                return (char) (value + 140);
            }
            return value;
        }

        public boolean isAlphanumeric(int value) {

            if (value == mezera || value == tab)
                return true;
            if (a <= value && value <= z)
                return true;
            if (A <= value && value <= Z)
                return true;
            if (zeroValue <= value && value <= nineValue)
                return true;

            return false;
        }
    }

}
