package os.session;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;
import java.util.logging.Logger;

import os.file.FileManager;
import os.history.History;
import os.history.HistoryManager;

/**
 * Manager for all session.
 * 
 * Create, login, logout, delete, cloning session.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 * 
 */
public class SessionManager {

    private static Logger logger = Logger.getLogger(SessionManager.class.getName());

    private static SessionManager manager;
    private static Long sessionIdGenerator = new Long(1);

    private ReentrantLock lock = new ReentrantLock();

    private Map<Long, Session> map = new HashMap<Long, Session>();
    private Map<String, Long> loginMap = new HashMap<String, Long>();

    public SessionManager() {

    }

    public static SessionManager get() {

        if (manager == null)
            init();

        return manager;
    }

    private static void init() {
        logger.log(Level.INFO, "initializing Session manager \n");
        manager = new SessionManager();
        logger.log(Level.INFO, "initialized Session manager \n");
    }

    /**
     * Shutdown session manager - logout all active session.
     */
    public void shutdown() {

        lock.lock();
        logger.log(Level.INFO, "shutdown Session manager \n");

        // all logged session have to logout on shutdown.
        for (Session session : map.values()) {

            if (session.isLogged()) {
                logger.log(Level.INFO, "logout session " + session.getLogin());
                HistoryManager.get().saveHistory(session);
                session.setLogin(null);
                session.setLogged(false);
            }

        }
        lock.unlock();
    }

    public Session getSession(Long id) {
        return map.get(id);
    }

    /**
     * Create new notLogged session.
     */
    public Session getNewSession() {

        Session session = new Session();

        lock.lock();

        session.setId(sessionIdGenerator++);
        map.put(session.getId(), session);

        lock.unlock();

        String dir = FileManager.get().getDefaultDir();
        session.setDirectory(dir);

        return session;
    }

    /**
     * Get copy of session with unique session ID.
     */
    public Session getCopyOfSession(Session origin) {

        Session session = new Session();
        session.setLogin(origin.getLogin());
        session.setLogged(origin.isLogged());
        session.setDirectory(origin.getDirectory());
        session.setHistory(origin.getHistory());

        lock.lock();

        session.setId(sessionIdGenerator++);
        map.put(session.getId(), session);

        lock.unlock();

        return session;

    }

    /**
     * Login session with name and prepare history if exist.
     */
    public void loginSession(Session session, String name) {

        // login session and load history
        String login = name.toLowerCase();
        logger.log(Level.INFO, "login session " + login);
        Long existSessionId = loginMap.get(login);
        // if not exist create new one
        if (existSessionId == null) {
            session.setLogin(login);
            session.setLogged(true);

            lock.lock();
            loginMap.put(session.getLogin(), session.getId());
            lock.unlock();

            // try load or find exist history for login
            History history = HistoryManager.get().getHistoryForSession(session);
            session.setHistory(history);

        } else {
            // or clone exist
            lock.lock();
            Session existSession = map.get(existSessionId);
            lock.unlock();

            session.setHistory(existSession.getHistory());
            session.setDirectory(existSession.getDirectory());
            session.setLogin(login);
            session.setLogged(true);
        }

    }

    /**
     * Logout session and save history.
     */
    public void logoutSession(Session session) {

        // save history and logout session
        logger.log(Level.INFO, "logout session " + session.getLogin());

        lock.lock();

        map.remove(session.getId());
        loginMap.remove(session.getLogin());

        lock.unlock();

        HistoryManager.get().saveHistory(session);
        session.setLogin(null);
        session.setLogged(false);

    }

}
