package os.terminal;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.io.IOException;
import java.io.PipedReader;
import java.io.PipedWriter;

import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;
import javax.swing.text.DefaultCaret;

import os.session.Session;

/**
 * Terminal tab in window.
 * 
 * @author Jakub Rinkes, Jan Moulis, Kamil Rendl
 * 
 */
public class TerminalTab extends JScrollPane {

    private static final String ERROR = "ERROR: ";
    private static final long serialVersionUID = -5321896386074814305L;
    private JTextArea inputArea;
    private JTextArea outputArea;
    private TerminalKeyTabListener keyListener;

    private boolean isWrited = false;

    public TerminalTab() {

        Dimension windowSize = new Dimension(800, 500);
        Font font = new Font("Consolas", Font.PLAIN, 12);

        outputArea = new JTextArea() {

            private static final long serialVersionUID = -1094612281359185309L;

            @Override
            public boolean isVisible() {
                return !getText().isEmpty();
            }

        };
        outputArea.setColumns(80);
        outputArea.setEditable(false);
        outputArea.setFocusable(false);
        outputArea.setLineWrap(true);
        outputArea.setWrapStyleWord(true);
        outputArea.setFont(font);
        // alway update caret for scrolling down via append text
        DefaultCaret caret = (DefaultCaret) outputArea.getCaret();
        caret.setUpdatePolicy(DefaultCaret.ALWAYS_UPDATE);

        inputArea = new JTextArea();
        keyListener = new TerminalKeyTabListener(outputArea, inputArea);
        inputArea.addKeyListener(keyListener);
        inputArea.setColumns(80);
        inputArea.setRows(1);
        inputArea.setLineWrap(true);
        inputArea.setWrapStyleWord(true);
        inputArea.setFont(font);
        // alway update caret for scrolling down via append text
        DefaultCaret caret2 = (DefaultCaret) inputArea.getCaret();
        caret2.setUpdatePolicy(DefaultCaret.ALWAYS_UPDATE);

        // magic layout for command line look - do not change this.
        JPanel panel = new JPanel();
        panel.setLayout(new BorderLayout());
        panel.add(outputArea, BorderLayout.NORTH);
        JPanel inner2 = new JPanel(new BorderLayout());
        inner2.add(inputArea, BorderLayout.NORTH);
        panel.add(inner2, BorderLayout.CENTER);
        requestFocusInWindow(); // this panel always get focus on click in window.

        // set scroll bar policy
        setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
        setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        setMinimumSize(windowSize);
        setViewportView(panel);

        // always scroll down
        getVerticalScrollBar().addAdjustmentListener(new AdjustmentListener() {
            public void adjustmentValueChanged(AdjustmentEvent e) {
                if (isWrited) {
                    isWrited = false;
                    e.getAdjustable().setValue(e.getAdjustable().getMaximum());
                }
            }
        });
    }

    public void writeOutput(final String text) {

        SwingUtilities.invokeLater(new Runnable() {

            public void run() {
                isWrited = true;
                outputArea.append(text);
                updateUI();
            }
        });
    }

    public void writeError(final String text) {

        // swing thread for update swing component - faster and not blocking awt thread
        SwingUtilities.invokeLater(new Runnable() {

            public void run() {
                isWrited = true;
                String output = ERROR + text;
                outputArea.append(output);
                updateUI();
            }
        });

    }

    public void connectInput(PipedReader reader) {

        try {
            isWrited = true;
            PipedWriter writer = keyListener.getWriter();
            writer.connect(reader);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void close() throws IOException {
        inputArea.setEditable(false);
        keyListener.getWriter().close();
    }

    public void setSession(Session session) {
        keyListener.setSession(session);
    }

}
